# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: caffe2/proto/predictor_consts.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='caffe2/proto/predictor_consts.proto',
  package='caffe2',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=b'\n#caffe2/proto/predictor_consts.proto\x12\x06\x63\x61\x66\x66\x65\x32\"\xa3\x05\n\x0fPredictorConsts\x12$\n\x0cMETA_NET_DEF\x18\x01 \x01(\t:\x0e!!META_NET_DEF\x12\x30\n\x12PREDICTOR_DBREADER\x18\x02 \x01(\t:\x14!!PREDICTOR_DBREADER\x12\x32\n\x14PARAMETERS_BLOB_TYPE\x18\x03 \x01(\t:\x14PARAMETERS_BLOB_TYPE\x12*\n\x10INPUTS_BLOB_TYPE\x18\x04 \x01(\t:\x10INPUTS_BLOB_TYPE\x12,\n\x11OUTPUTS_BLOB_TYPE\x18\x05 \x01(\t:\x11OUTPUTS_BLOB_TYPE\x12\x32\n\x14GLOBAL_INIT_NET_TYPE\x18\x06 \x01(\t:\x14GLOBAL_INIT_NET_TYPE\x12\x34\n\x15PREDICT_INIT_NET_TYPE\x18\x07 \x01(\t:\x15PREDICT_INIT_NET_TYPE\x12*\n\x10PREDICT_NET_TYPE\x18\x08 \x01(\t:\x10PREDICT_NET_TYPE\x12*\n\x10SINGLE_PREDICTOR\x18\t \x01(\t:\x10SINGLE_PREDICTOR\x12(\n\x0fMULTI_PREDICTOR\x18\n \x01(\t:\x0fMULTI_PREDICTOR\x12\x32\n\x14TRAIN_INIT_PLAN_TYPE\x18\x0b \x01(\t:\x14TRAIN_INIT_PLAN_TYPE\x12(\n\x0fTRAIN_PLAN_TYPE\x18\x0c \x01(\t:\x0fTRAIN_PLAN_TYPE\x12(\n\x0fSHAPE_INFO_BLOB\x18\r \x01(\t:\x0fSHAPE_INFO_BLOB\x12\x36\n\x14\x44\x45\x46\x45RRED_BLOB_READER\x18\x0e \x01(\t:\x18__DEFERRED_BLOB_READER__'
)




_PREDICTORCONSTS = _descriptor.Descriptor(
  name='PredictorConsts',
  full_name='caffe2.PredictorConsts',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='META_NET_DEF', full_name='caffe2.PredictorConsts.META_NET_DEF', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=b"!!META_NET_DEF".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='PREDICTOR_DBREADER', full_name='caffe2.PredictorConsts.PREDICTOR_DBREADER', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=b"!!PREDICTOR_DBREADER".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='PARAMETERS_BLOB_TYPE', full_name='caffe2.PredictorConsts.PARAMETERS_BLOB_TYPE', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=b"PARAMETERS_BLOB_TYPE".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='INPUTS_BLOB_TYPE', full_name='caffe2.PredictorConsts.INPUTS_BLOB_TYPE', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=b"INPUTS_BLOB_TYPE".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='OUTPUTS_BLOB_TYPE', full_name='caffe2.PredictorConsts.OUTPUTS_BLOB_TYPE', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=b"OUTPUTS_BLOB_TYPE".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='GLOBAL_INIT_NET_TYPE', full_name='caffe2.PredictorConsts.GLOBAL_INIT_NET_TYPE', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=b"GLOBAL_INIT_NET_TYPE".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='PREDICT_INIT_NET_TYPE', full_name='caffe2.PredictorConsts.PREDICT_INIT_NET_TYPE', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=b"PREDICT_INIT_NET_TYPE".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='PREDICT_NET_TYPE', full_name='caffe2.PredictorConsts.PREDICT_NET_TYPE', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=b"PREDICT_NET_TYPE".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='SINGLE_PREDICTOR', full_name='caffe2.PredictorConsts.SINGLE_PREDICTOR', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=b"SINGLE_PREDICTOR".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='MULTI_PREDICTOR', full_name='caffe2.PredictorConsts.MULTI_PREDICTOR', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=b"MULTI_PREDICTOR".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='TRAIN_INIT_PLAN_TYPE', full_name='caffe2.PredictorConsts.TRAIN_INIT_PLAN_TYPE', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=b"TRAIN_INIT_PLAN_TYPE".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='TRAIN_PLAN_TYPE', full_name='caffe2.PredictorConsts.TRAIN_PLAN_TYPE', index=11,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=b"TRAIN_PLAN_TYPE".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='SHAPE_INFO_BLOB', full_name='caffe2.PredictorConsts.SHAPE_INFO_BLOB', index=12,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=b"SHAPE_INFO_BLOB".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='DEFERRED_BLOB_READER', full_name='caffe2.PredictorConsts.DEFERRED_BLOB_READER', index=13,
      number=14, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=b"__DEFERRED_BLOB_READER__".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=48,
  serialized_end=723,
)

DESCRIPTOR.message_types_by_name['PredictorConsts'] = _PREDICTORCONSTS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

PredictorConsts = _reflection.GeneratedProtocolMessageType('PredictorConsts', (_message.Message,), {
  'DESCRIPTOR' : _PREDICTORCONSTS,
  '__module__' : 'caffe2.proto.predictor_consts_pb2'
  # @@protoc_insertion_point(class_scope:caffe2.PredictorConsts)
  })
_sym_db.RegisterMessage(PredictorConsts)


# @@protoc_insertion_point(module_scope)
