#pragma once

#include <memory>
#include <vector>

#include "test/cpp/tensorexpr/test_base.h"
#include "torch/csrc/jit/testing/file_check.h"

namespace torch {
namespace jit {
using namespace torch::jit::tensorexpr;

} // namespace jit
} // namespace torch
